require 'uu/os/qos/qos_handler'
require 'uu/os/rest/remote_client'
require 'uu/os/uesuri'
require 'uu/os/uesuri_builder'
require 'uu/os/attachment'
require 'uu/os/search'
require 'uu/os/use_case_realization'
require 'uu/os/security/session'
require 'uu/os/vuc/http_status'
require 'uu/os/vuc/view'
require 'uu/os/logger'
require 'uu/os/vuc/logger_context'
require 'uu/os/vuc/use_case_log_writer'

module UU
  module OS
    module VUC

      # Class for dispatching controllers stored as attachments. All controllers should define only event
      # handling methods. It is not recommended for controllers stored as attachments to inherit from any
      # class (including {UU::OS::VUC::FormController}).
      #
      # @example Event method definition:
      #           def on_<%event_code%>(view)
      #             # 1. process and modify view
      #             # 2. return status from {UU::OS::VUC::HTTPStatus}
      #             #    (optional, 200/OK is returned if not defined)
      #           end
      class AttachmentDispatcher

        # Cache holding versions of loaded controllers.
        # @private
        @@controller_version = {}

        # Instance of logger.
        # @private
        @@log = UU::OS::Logger.new(self)

        # Dispatches functionality of controller stored as attachment on
        # form artifact. Particular controller is loaded based on data
        # in +body+ parameter, which contains view model send by form view.
        #
        # @param [String] base_url Base URL on which was controller called
        # @param [String] auth_token Authentication token to be used by default
        # @param [String] body Request body containing view data
        #
        # @return [Array<Numeric, String>] Array containing HTTP status code (should be one of {UU::OS::VUC::HTTPStatus}) as first element and response body as second element
        #
        # @private
        def self.dispatch(base_url, auth_token, body)
          ctrl_cls = nil
          begin
            # Controllers should respond as fast as possible, they should not wait on QoS
            UU::OS::QoS::QoSHandler.default_max_client_timeout = 0
            # If not configured, call all services on same base URL as this controller
            UU::OS::REST::RemoteClient.default_base_url = base_url
            # Parse incomming request and create view model
            view = UU::OS::VUC::View.new(body)
            # Set diagnostic information to logger context
            UU::OS::VUC::LoggerContext.set(view.context)
            # If authentication data are provided, authenticate user
            # (Authentication is called before every other operation to be sure no
            # processing is done without valid authentication data)
            UU::OS::Security::Session.send(:init_session, auth_token) if auth_token
            # Load and initialize controller from attachment
            ctrl_cls = load_controller(view)
            if !ctrl_cls
              raise 'Controller attachment does not exist or does not contain valid controller implementation.'
            end
            ctrl = ctrl_cls.new
            if ctrl.class.name == 'Sinatra::Wrapper'
              # In case controller in attachment was inherited from Sinatra (either directly or
              # via FormController class) we have to read actual instance from inner variable.
              ctrl = ctrl.instance_variable_get(:@instance)
            end
            # Prepare expected event method
            ctrl_method = get_event_method(view.event.code)
            do_log = (ctrl_method =~ /^on_init$/)
            if do_log
              log_params = {}
              log_params[:start] = Time.now
              log_params[:level] = UU::OS::UseCaseLog::UseCaseLogLevel::ERROR
              log_params[:type] = UU::OS::UseCaseLog::UseCaseLogType::VISUAL_USE_CASE
              log_params[:session] = UU::OS::Security::Session.current_session
              log_params[:view] = view
            end
            # Check if controller defines invoked event method
            if !ctrl.respond_to?ctrl_method
              if ctrl_method == :on_submit
                # In case method "submit" is not defined, throw exception - such controller is not
                # able to handle forms at all.
                raise "Controller #{ctrl.class} does not define on_submit method."
              end
              # Unimplemented event should not cause any problem except "submit" which is handled above
              if do_log
                log_params[:level] = UU::OS::UseCaseLog::UseCaseLogLevel::INFO
              end
              resp_body = view.to_response
              log_request(ctrl.class, ctrl_method, body, resp_body)
              return [HTTPStatus::OK, resp_body]
            end
            
            # TODO Parse view again, to allow loaded controller to inject additional component types.
            # e.g. uu_datatable - Controller requires gem uu_databable, which injects new component
            # type. This type is used only when parsing view, but view was parsed before loading of
            # controller (to be able to find proper controller class). Therefore we are parsing view
            # once more. This code should be optimized, e.g. use only partial parse of view to load
            # controller.
            view = UU::OS::VUC::View.new(body)
            # Invoke controller method according to incoming event
            ctrl.send(ctrl_method, view)
            status = view.status
            if (status.nil?) || (!status.kind_of?Numeric)
              # Invoke method did not provide status code or
              # returned some other object than status number
              status = HTTPStatus::OK
            end
            if do_log
              log_params[:level] = UU::OS::UseCaseLog::UseCaseLogLevel::INFO if status.to_i < 400
            end
            resp_body = view.to_response
            log_request(ctrl.class, ctrl_method, body, resp_body)
            [status, resp_body]
          rescue Exception => e
            log_params[:message] = e.to_s if do_log
            get_logger(ctrl_cls).error('Execution of controller failed.', e)
            [HTTPStatus::INTERNAL_SERVER_ERROR, get_error_response(e)]
          ensure
            if do_log
              log_params[:finish] = Time.now
              UU::OS::VUC::UseCaseLogWriter.add_record(log_params)
            end
            # Always do complete deletion of all tokens
            UU::OS::Security::Session.logout(true)
            # Always clean up all diagnostic information from logger context
            UU::OS::VUC::LoggerContext.clean
          end
        end

        # Loads controller (use case realization class) from attachment
        # based on data from given View object.
        #
        # @param [UU::OS::VUC::View] view View object send by form view containing controller definition
        #
        # @return [Class] Controller class
        #
        # @private
        def self.load_controller(view)
          # Load use case realization
          ucr_uri = UU::OS::UESURI.new(view.context.use_case_realization_uri)
          # Try to get controller class
          controller = get_class_by_name(view.context.controller_class)
          if controller
            if  view.context.controller_revision.to_i <= @@controller_version[controller].to_i
              # If controller is loaded and is in most recent version, return it
              return controller
            end
            # If controller is older, unload its class from memory
            unload_controller(controller)
          end
          # Load controller data
          data = UU::OS::UseCaseRealization.get_data(ucr_uri)
          # Name temporary attachment file according to use case (file is displayed in stacktrace)
          tmp = Tempfile.new("controller-#{get_controller_filename(ucr_uri.artifact_code)}-")
          begin
            if data.data.respond_to?(:read)
              tmp.write(data.data.read.force_encoding('UTF-8'))
            else
              tmp.write(data.data)
            end
            tmp.rewind
            load tmp.path
          ensure
            tmp.close()
            tmp.unlink()
          end
          # Try to get controller class (now it should be loaded)
          controller = get_class_by_name(view.context.controller_class)
          if controller
            # Store version of loaded controller
            @@controller_version[controller] = view.context.controller_revision
          end
          controller
        end
        private_class_method :load_controller

        # Unloads controller class from memory. This method is used to be able to load
        # new version of controller from attachment.
        #
        # @param [Class] controller Controller class to be unloaded
        #
        # @private
        def self.unload_controller(controller)
          parent = get_parent_module(controller)
          cls_name = controller.name.reverse.split('::', 2)[0].reverse
          if parent.const_defined?cls_name
            parent.send(:remove_const, cls_name)
          end
        end
        private_class_method :unload_controller

        # Returns instance of +Class+ object for given class name
        #
        # @param [String] name Class name
        #
        # @return [Class] Instance of +Class+ object or nil if no such class exists
        #
        # @private
        def self.get_class_by_name(name)
          name.split('::').inject(Object) do |mod, name|
            if (mod.const_defined?(name))
              mod.const_get(name)
            else
              return nil
            end
          end
        end
        private_class_method :get_class_by_name

        # Returns parent module of given +Class+ object.
        #
        # @param [Class] cls Instance of +Class+ object
        #
        # @return [Class] Parent module +Class+ object
        #
        # @private
        def self.get_parent_module(cls)
          parts = cls.name.reverse.split('::', 2)
          if parts.size == 2
            get_class_by_name(parts[1].reverse)
          else
            Object
          end
        end
        private_class_method :get_parent_module

        # Returns valid filename based on controller class name.
        #
        # @param [String] name Name to be transformed to valid filename
        #
        # @return [String] Valid filename
        #
        # @private
        def self.get_controller_filename(name)
          filename = name.to_s.dup
          filename.gsub!(/[^\w\s_-]+/, '_')
          filename.gsub!(/(^|\b\s)\s+($|\s?\b)/, '\\1\\2')
          filename.gsub!(/\s+/, '_')
          filename
        end
        private_class_method :get_controller_filename

        # Returns name of method for handling of event
        # (transforms camel case to snake case).
        #
        # @param [String] event Event to be handled
        #
        # @return [Symbol] Name of method for handling event
        #
        # @private
        def self.get_event_method(event)
          # Based on ActiveSupport::Inflector.underscore
          # TODO This method is also on FormController. We need some "util" class/module to share this code
          result = event.to_s.dup
          result.gsub!(/([A-Z]+)([A-Z][a-z])/,'\1_\2')
          result.gsub!(/([a-z\d])([A-Z])/,'\1_\2')
          result.tr!('-', '_')
          ('on_' << result.downcase).to_sym
        end
        private_class_method :get_event_method

        # Returns form error response.
        #
        # @param [Exception] exception Instance of exception
        #
        # @return [String] Error response
        #
        # @private
        def self.get_error_response(exception)
          # TODO This method is also on FormController. We need some "util" class/module to share this code
          id = (exception.respond_to?:id) ? exception.id : ''
          code = (exception.respond_to?:code) ? exception.code : ''
          message = (exception.respond_to?:message) ? exception.message : ''
          "{\"id\":\"#{id}\",\"code\":\"#{code}\",\"message\":\"#{message}\",\"time\":\"#{Time.now}\"}"
        end
        private_class_method :get_error_response

        # Returns instance of logger.
        #
        # @param [String, Class] clazz Controller class (may be +nil+)
        #
        # @return [UU::OS::Logger] Instance of logger
        #
        # @private
        def self.get_logger(clazz = nil)
          if clazz.nil?
            return @@log
          else
            return UU::OS::Logger.new(clazz)
          end
        end
        private_class_method :get_logger

        # Logs form request
        #
        # @param [String, Class] clazz Controller class
        # @param [String, Symbol] method Invoked method
        # @param [String] request Incoming request
        # @param [String] response Outgoing response
        #
        # @private
        def self.log_request(clazz, method, request, response)
          log = get_logger(clazz)
          if log.debug_loggable?
            req = (request.nil?) ? request : request.dup.force_encoding('UTF-8')
            resp = (response.nil?) ? response : response.dup.force_encoding('UTF-8')
            log.debug("Request processed.\n\tController: #{clazz}\n\tMethod: #{method}\n\tRequest: #{req}\n\tResponse:#{resp}")
          end
        end
        private_class_method :log_request

      end

    end
  end
end
