require 'json'

module UU
  module OS
    module VUC

      # Use case context object.
      class Context

        # Raw context attributes.
        # @private
        attr_reader :raw_attrs

        # URI of use case referrer. Value may be +nil+ (e.g. in case of session-free-link navigation).
        # @return [String]
        attr_reader :referrer_uri

        # UESURI of use case.
        # @return [String]
        attr_reader :use_case_uri

        # UESURI of use case realization.
        # @return [String]
        attr_reader :use_case_realization_uri

        # UESURI of application instance.
        # @return [String]
        attr_reader :application_instance_uri

        # URL where application instance is deployed.
        # @return [String]
        attr_reader :application_instance_url_path

        # Use case execution ID.
        # @return [String]
        attr_reader :execution_id

        # URL where application controller is deployed.
        # @return [String]
        attr_reader :controller_url_path

        # Controller realization class.
        # @return [String]
        attr_reader :controller_class

        # Controller realization version.
        # @return [String]
        attr_reader :controller_revision

        # UESURI of main entity.
        # @return [String]
        attr_reader :main_entity_uri

        # UESURI of territory.
        # @return [String]
        attr_reader :territory_uri

        # UESURI of access role.
        # @return [String]
        attr_reader :access_role_uri

        # UESURI of executive role.
        # @return [String]
        attr_reader :executive_role_uri

        # Authentication token.
        # @return [String]
        attr_reader :jwt_token

        # Use case parameters.
        # @return [Hash]
        attr_reader :parameters

        # Attributes to be passed between control methods of VUC (within single UC execution).
        # @return [Hash]
        attr_reader :attributes

        # Creates new instance of use case context.
        #
        # @param [String, Hash] data Initialization data in form of JSON string or Hash (optional)
        def initialize(data = nil)
          if data.nil?
            @raw_attrs = {}
            return
          end
          if data.kind_of?String
            data = JSON.parse(data, :symbolize_names => true)
          end
          if data.kind_of?Hash
            @raw_attrs = data
          elsif data.kind_of?Context
            if data.raw_attrs
              @raw_attrs = data.raw_attrs
            else
              @raw_attrs = {}
            end
          else
            raise ArgumentError.new("Context data must be JSON String or Hash, but was #{data.class}.")
          end
        end

        def raw_attrs
          # Disallow manipulating with raw attributes
          @raw_attrs.clone if @raw_attrs
        end

        def referrer_uri
          @raw_attrs[:referrerUri]
        end

        def use_case_uri
          @raw_attrs[:useCaseUri]
        end

        def use_case_realization_uri
          @raw_attrs[:useCaseRealizationUri]
        end

        def application_instance_uri
          @raw_attrs[:applicationInstanceUri]
        end

        def application_instance_url_path
          @raw_attrs[:applicationInstanceUrlPath]
        end

        def execution_id
          @raw_attrs[:executionId]
        end

        def controller_url_path
          @raw_attrs[:controllerUrlPath]
        end

        def controller_class
          @raw_attrs[:controllerClass]
        end

        def controller_revision
          @raw_attrs[:controllerRevision]
        end

        def main_entity_uri
          @raw_attrs[:mainEntityUri]
        end

        def territory_uri
          @raw_attrs[:territoryUri]
        end

        def access_role_uri
          @raw_attrs[:accessRoleUri]
        end

        def executive_role_uri
          @raw_attrs[:executiveRoleUri]
        end

        def jwt_token
          @raw_attrs[:jwtToken]
        end

        def parameters
          # Disallow manipulating with parameters
          @raw_attrs[:params].clone if @raw_attrs[:params]
        end

        def attributes
          if @raw_attrs[:attributes].nil?
            @raw_attrs[:attributes] = {}
          end
          @raw_attrs[:attributes]
        end

      end

    end
  end
end
