require 'json'
require 'uu/os/vuc/event'
require 'uu/os/vuc/context'
require 'uu/os/vuc/navigation'
require 'uu/os/vuc/view_template'
require 'uu/os/gvc/container'
require 'uu/os/gvc/component'
require 'uu/os/gvc/message'

module UU
  module OS
    module VUC

      # View object used for communication with forms.
      class View < UU::OS::GVC::Container

        # Unique view ID.
        # @return [String]
        attr_reader :id

        # View code.
        # @return [String]
        attr_reader :code

        # View name.
        # @return [String]
        attr_reader :name

        # HTTP status code of processed request (possible values are available on {UU::OS::VUC::HTTPStatus}).
        # @param [Numeric]
        # @return [Numeric]
        attr_accessor :status

        # View type.
        # @return [Numeric]
        attr_reader :component_type

        # Incoming event.
        # @return [UU::OS::VUC::Event]
        attr_reader :event

        # Use case context.
        # @return [UU::OS::VUC::Context]
        attr_reader :context

        # What should form do next.
        # @return [UU::OS::VUC::Navigation]
        attr_reader :navigation

        # The result of the VUC. Setting the value to anything other than nil
        # means that the VUC has finished and the user can be navigated elsewhere.
        #
        # By convention, setting "false" means
        # that the VUC was cancelled. The result will be passed to the parent VUC
        # if the current VUC runs in a dialog window. Otherwise, if the value
        # is an URI, it'll be used for navigation. Otherwise, the user will be
        # navigated to the previous UC.
        attr_accessor :result
        
        # View template that can be used for setting different sheet that will be rendered as form. Read {UU::OS::VUC::ViewTemplate#data_uri} for more info. 
        # 
        # @note View template can be set only in on_before_init method, if you set data_uri in another method, it will have no effect. 
        # 
        # @example
        #   # uri of default VUC
        #   UCR_URI = "ues:UNI:CDS.GUESTBOOK/COMMENT/CREATE"
        #   
        #   # code of sheet with extended form
        #   EXTENDED_FORM_CODE = "EXTENDED_FORM"
        #   
        #   def on_before_init(ctx)
        #   
        #     # get default view template => main sheet on VUC
        #     view_template_uri = ctx.view_template.data_uri # String ues:UNI:CDS.GUESTBOOK/COMMENT/CREATE:DEFAULT_FORM
        #   
        #     use_extended_form = ... # some logic to find out if use can see extended form
        #   
        #     if use_extended_form
        #       # set new sheet with extended form
        #       ctx.view_template.data_uri = UU::OS::UESURIBuilder.parse_uesuri(UCR_URI).set_object_code(EXTENDED_FORM_CODE).to_uesuri
        #     end
        #   
        #     # if you do not set data_uri so default is used
        #   
        #   end
        # 
        # @see UU::OS::VUC::ViewTemplate
        # @return [UU::OS::VUC::ViewTemplate]
        attr_reader :view_template
        

        # Creates new instance of view.
        #
        # @param [String, Hash] data Initialization data in form of JSON string or Hash (optional)
        def initialize(data = nil)
          if data.nil?
            super(data)
            return
          end
          if data.kind_of?String
            data = JSON.parse(data, :symbolize_names => true)
          end
          if !data.kind_of?Hash
            raise ArgumentError.new("View data must be JSON String or Hash, but was #{data.class}.")
          end
          super(data)
          @id = data[:id]
          @code = data[:code]
          @name = data[:name]
          @status = data[:status]
          @component_type = data[:componentType]
          @event = UU::OS::VUC::Event.new(data[:event])
          @context = UU::OS::VUC::Context.new(data[:context])
          @navigation = UU::OS::VUC::Navigation.new(data[:navigation])
          @result = data[:result]
          @view_template =  UU::OS::VUC::ViewTemplate.new(data[:viewTemplate])
        end

        # Serializes view to JSON.
        #
        # @return [String] Serialized view
        def to_response
          response = {}
          response[:id] = @id
          response[:code] = @code
          response[:name] = @name
          response[:status] = @status
          response[:componentType] = @component_type
          response[:event] = @event.raw_attrs if (@event.raw_attrs) && (!@event.raw_attrs.empty?)
          response[:context] = @context.raw_attrs if (@context.raw_attrs) && (!@context.raw_attrs.empty?)
          if (@components) && (@components.values.size > 0)
            response[:components] = []
            @components.values.each do |component|
              response[:components] << component.attributes
            end
          end
          if (@messages) && (@messages.size > 0)
            response[:messages] = []
            @messages.each do |message|
              response[:messages] << message.raw_attrs
            end
          end
          response[:navigation] = @navigation.raw_attrs if (@navigation.raw_attrs) && (!@navigation.raw_attrs.empty?)
          response[:focusOn] = @focus_on if @focus_on
          response[:result] = @result if !@result.nil?
          response[:viewTemplate] = @view_template.raw_attrs if (@view_template.raw_attrs) && (!@view_template.raw_attrs.empty?)
          response.to_json
        end

      end

    end
  end
end
